﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Messaging;
using System.Text;
using System.Threading.Tasks;
using System.Xml.Linq;

namespace WebServerDocxGeneration
{
    class WebServerDocxGeneration
    {
        static string ControllerMachineName = Environment.MachineName.ToLower();
        static string DocGenMachineName = "bigi5-8";
        static int NumberOfDaemons = 4;
        static int DocumentsPerRequest = 500;

        static string ControllerQueueLocalName = @".\PRIVATE$\GenerateDocumentsController";
        static string ControllerQueueRemoteName = string.Format(@"FormatName:Direct=OS:{0}\PRIVATE$\GenerateDocumentsController", ControllerMachineName);

        static MessageQueue ControllerQueue;

        public static FileInfo TemplateDocumentFileInfo = new FileInfo(@"\\bigi5-8\bigi5-Z\TemplateDocuments\TemplateDocument.docx");
        static DirectoryInfo GeneratedDocumentsDirectory = new DirectoryInfo(@"\\bigi5-8\bigi5-Z\GeneratedDocuments\");

        static List<string> RemoteQueues;

        static void Main(string[] args)
        {
            Initialize();
            GenerateDocuments();
        }

        /*
         * Outgoing message:
         * 
         * Message label: GenerateDocuments
         * 
         * <GenerateDocuments>
         *   <ControllerQueue Name=".\PRIVATE$\GenerateDocumentController" />
         *   <TemplateDocument Name="C:\Users\Public\Documents\TemplateDocuments\TemplateDocument.docx" />
         *   <GeneratedDocumentsDirectory Name="C:\Users\Public\Documents\GeneratedDocuments" />
         *   <DocumentGenerationData>
         *     <!-- the XML here is specific to the template document.  If the template document changes, the schema of the following will change to match it. -->
         *     <Customers>
         *       <Customer>
         *         <CustomerID>1</CustomerID>
         *         <Name>Cheryl</Name>
         *         <HighValueCustomer>True</HighValueCustomer>
         *         <Orders>
         *           <Order>
         *             <ProductDescription>Unicycle</ProductDescription>
         *             <Quantity>3</Quantity>
         *             <OrderDate>9/5/2001</OrderDate>
         *           </Order>
         *           <Order>
         *             <ProductDescription>Tricycle</ProductDescription>
         *             <Quantity>3</Quantity>
         *             <OrderDate>8/6/2000</OrderDate>
         *           </Order>
         *         </Orders>
         *       </Customer>
         *     </Customers>
         *   </DocumentGenerationData>
         * </GenerateDocuments>
         */

        /*
         * Return message:
         * 
         * Message label: DocumentGenerationComplete
         * 
         * <DocumentsGenerated>
         *   <GenerateDocumentsQueue Name="FormatName:Direct=OS:mini-1\PRIVATE$\GenerateDocuments1" />
         *   <Document Name="48FCE74D-E9E1-49DB-ABCD-629196B6AEE5.docx" />
         * </DocumentsGenerated>
         */
        private static void GenerateDocuments()
        {
            int documentsToGenerate = 30000;
            int documentsRequested = 0;
            int totalDocumentsGenerated = 0;
            DateTime startTime = DateTime.Now;

            while(true)
            {
                Message controllerMessage = new Message();

                controllerMessage = ControllerQueue.Receive();
                controllerMessage.Formatter = new XmlMessageFormatter(new String[] { "System.String,mscorlib" });
                var controllerMessageText = controllerMessage.Body.ToString();

                if (controllerMessage.Label == "DocumentGenerationComplete")
                {
                    var generateDocumentsArguments = XElement.Parse(controllerMessageText);

                    // If you want to validate the XML in the message, this is the place to do it.

                    totalDocumentsGenerated += generateDocumentsArguments.Elements("Document").Count();
                    TimeSpan elapsedTime = DateTime.Now - startTime;
                    var docsPerMinute = (int)((double)totalDocumentsGenerated / elapsedTime.TotalMinutes);
                    Console.WriteLine("Documents generated: {0}  Documents per minute: {1}", totalDocumentsGenerated, docsPerMinute);
                    if (totalDocumentsGenerated == 30000)
                    {
                        Console.WriteLine("Press Enter");
                        Console.ReadKey();
                        Environment.Exit(0);
                    }
                }
                if (controllerMessage.Label == "DocumentGenerationComplete" || controllerMessage.Label == "Ready")
                {
                    if (documentsRequested < documentsToGenerate)
                    {
                        var controllerMessageArguments = XElement.Parse(controllerMessageText);

                        // If you want to validate the XML in the message, this is the place to do it.

                        var genDocRemoteQueueName = (string)controllerMessageArguments
                            .Element("GenerateDocumentsQueue")
                            .Attribute("Name");

                        SendGenerateDocumentMessage(genDocRemoteQueueName);
                        documentsRequested += DocumentsPerRequest;
                    }
                    continue;
                }
                else
                    Console.WriteLine("Unexpected message");
            }
        }

        private static void SendGenerateDocumentMessage(string genDocRemoteQueueName)
        {
            MessageQueue remoteQueue = new MessageQueue(genDocRemoteQueueName, false, false, QueueAccessMode.Send);
            Message generateDocumentsMessage = new Message();
            var gmsg = new XElement("GenerateDocuments",
                new XElement("ControllerQueue", new XAttribute("Name", ControllerQueueRemoteName)),
                new XElement("TemplateDocument", new XAttribute("Name", TemplateDocumentFileInfo.FullName)),
                new XElement("GeneratedDocumentsDirectory", new XAttribute("Name", GeneratedDocumentsDirectory.FullName)),
                new XElement("DocumentGenerationData",
                    new XElement("Customers",
                        Enumerable.Range(1, DocumentsPerRequest).Select(r =>
                            new XElement("Customer",
                                new XElement("CustomerID", 1),
                                new XElement("Name", "Cheryl"),
                                new XElement("HighValueCustomer", true),
                                new XElement("Orders",
                                    new XElement("Order",
                                        new XElement("ProductDescription", "Unicycle"),
                                        new XElement("Quantity", 3),
                                        new XElement("OrderDate", "9/5/2001")),
                                    new XElement("Order",
                                        new XElement("ProductDescription", "Tricycle"),
                                        new XElement("Quantity", 3),
                                        new XElement("OrderDate", "9/5/2001"))))))));

            generateDocumentsMessage.Body = gmsg.ToString(SaveOptions.DisableFormatting);
            generateDocumentsMessage.Label = "GenerateDocuments";
            remoteQueue.Send(generateDocumentsMessage);
        }

        private static void Initialize()
        {
            RemoteQueues = Enumerable.Range(1, NumberOfDaemons)
                .Select(d => string.Format(@"FormatName:Direct=OS:{0}\PRIVATE$\GenerateDocuments{1}", DocGenMachineName, d))
                .ToList();

            if (!GeneratedDocumentsDirectory.Exists)
                GeneratedDocumentsDirectory.Create();

            if (MessageQueue.Exists(ControllerQueueLocalName))
                MessageQueue.Delete(ControllerQueueLocalName);
            ControllerQueue = MessageQueue.Create(ControllerQueueLocalName, false);

            foreach (var genDocQueue in RemoteQueues)
            {
                MessageQueue remoteQueue = new MessageQueue(genDocQueue, false, false, QueueAccessMode.Send);
                Message pingMessage = new Message();
                var pingMsg = new XElement("Ping",
                    new XElement("ControllerQueue",
                        new XAttribute("Name", ControllerQueueRemoteName)));
                pingMessage.Body = pingMsg.ToString(SaveOptions.DisableFormatting);
                pingMessage.Label = "Ping";
                remoteQueue.Send(pingMessage);
            }
        }
    }
}
